from transformers import AutoModelForCausalLM, AutoTokenizer,AutoModelForPreTraining,BitsAndBytesConfig
import torch
from typing import Optional, Callable
from torch import nn
from transformers import AutoModelForCausalLM
import types
import os
from transformers.models.bloom.modeling_bloom import BloomForCausalLM
from .kv_cache_model import KVCacheModel
from .speculative_decode import *
from .generate import Generate

class GPTFast:

    def __model_compile(model:nn.Module, spec_decode:bool = False) -> nn.Module:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        model = torch.compile(model, mode="reduce-overhead", fullgraph=True)
        model.prefill = torch.compile(model.prefill, dynamic=True, fullgraph=True)
        model = model.to(device)
        if spec_decode:
            assert hasattr(model, "draft_model"), "You have passed spec_decode = True in your torch_compile but your model doesn't have a draft model."
            draft_model = model.draft_model
            draft_model = torch.compile(draft_model, mode="reduce-overhead", fullgraph=True)
            draft_model.prefill = torch.compile(draft_model.prefill, dynamic=True, fullgraph=True)
            draft_model = draft_model.to(device)

        return model

    def __gpt_fast(model_name:str,trust_remote_code=False, **spec_dec_kwargs):
        quantization_config = BitsAndBytesConfig(load_in_8bit=True)
        try:
            model = AutoModelForCausalLM.from_pretrained(model_name, quantization_config = quantization_config,trust_remote_code=trust_remote_code)
        except:
            model = AutoModelForPreTraining.from_pretrained(model_name, quantization_config = quantization_config,trust_remote_code=trust_remote_code)
        
        model = KVCacheModel(model)
        spec_decode = False
        if spec_dec_kwargs:
            draft_model_name = spec_dec_kwargs.pop("draft_model_name")
            try:
                draft_model = AutoModelForCausalLM.from_pretrained(draft_model_name,trust_remote_code=trust_remote_code)
            except:
                draft_model = AutoModelForPreTraining.from_pretrained(draft_model_name,trust_remote_code=trust_remote_code)
            draft_model = KVCacheModel(draft_model)
            model = add_speculative_decoding(model, draft_model, **spec_dec_kwargs)
            spec_decode = True
        model = GPTFast.__model_compile(model, spec_decode=spec_decode)
        return model
    
    def __argmax(self, probabilities):
        # Use argmax to get the token with the maximum probability
        max_prob_index = torch.argmax(probabilities, dim=-1)
        return max_prob_index.unsqueeze(0)

    def __generate_output_fast(model_name, draft_model_name,sequence, max_length,trust_remote_code=False):
        print("Experimental!!! Generating output using gpt fast with speculative decoding upto 6x faster than normal gpt generation. if fails go back to normal gpt generation.")
        device = "cuda" if torch.cuda.is_available() else "cpu"
        model = GPTFast.__gpt_fast(model_name, trust_remote_code=trust_remote_code, draft_model_name=draft_model_name, sample_function=GPTFast.__argmax)
        model.to(device)
        tokenizer = Generate.load_tokenizer(model_name)
        input_tokens = tokenizer.encode(sequence, return_tensors="pt").to(device)
        output = model.generate(cur_tokens=input_tokens, max_tokens=max_length, speculate_k=6)
        return output,tokenizer
    
    def __generate_output_from_model(model, tokenizer, sequence, max_length):
        print("Experimental!!! Generating output using gpt fast with speculative decoding upto 6x faster than normal gpt generation. if fails go back to normal gpt generation.")
        device = "cuda" if torch.cuda.is_available() else "cpu"
        input_tokens = tokenizer.encode(sequence, return_tensors="pt").to(device)
        output = model.generate(cur_tokens=input_tokens, max_tokens=max_length, speculate_k=6)
        return output,tokenizer
    
    def gpt_fast(model_name:str ,trust_remote_code=False, **spec_dec_kwargs):
        return GPTFast.__gpt_fast(model_name, trust_remote_code=trust_remote_code, **spec_dec_kwargs)
    
    def generate_output_fast(model_name, draft_model_name, sequence, max_length, trust_remote_code=False):
        return GPTFast.__generate_output_fast(model_name, draft_model_name, sequence, max_length, trust_remote_code=trust_remote_code)
    
    def generate_text_fast(model_name, draft_model_name, sequence, max_length, trust_remote_code=False):
        output,tokenizer = GPTFast.__generate_output_fast(model_name, draft_model_name, sequence, max_length, trust_remote_code=trust_remote_code)
        print(tokenizer.decode(output[0], skip_special_tokens=True))
        return tokenizer.decode(output[0], skip_special_tokens=True)
    
    def load_model(model_path,draft_model_name,trust_remote_code=False):
        return GPTFast.__gpt_fast(model_path, trust_remote_code=trust_remote_code, draft_model_name=draft_model_name, sample_function=GPTFast.__argmax)
    
    def generate_output_from_model(model, tokenizer, sequence, max_length):
        return GPTFast.__generate_output_from_model(model, tokenizer, sequence, max_length)
    


import torch
from torch import nn
from typing import Optional, Callable
from transformers.models.bloom.modeling_bloom import BloomForCausalLM

class KVCacheModel(nn.Module):
    def __init__(self, model : torch.nn.Module) -> None:
        super().__init__()
        self._model = model
        self._past_key_values = None
        self._prob_history = None
        self._vocab_size = self._model.config.vocab_size

    @torch.no_grad()
    def prefill(self, input_ids:torch.Tensor) -> torch.Tensor:
        assert self._past_key_values is None and self._prob_history is None
        outputs = self._model(input_ids)
        self._prob_history = outputs.logits #(batch_size, seq_len, vocab_size)
        self._past_key_values = outputs.past_key_values
        last_q = self._prob_history[:, -1, :]
        return last_q

    @torch.no_grad()
    def forward(self, input_ids:torch.Tensor) -> torch.Tensor:
        # return the last token's logits
        k, v = self._past_key_values[-1]
        cached_len = k.shape[2]

        last_input_id = input_ids[:, cached_len:]

        outputs = self._model(last_input_id, past_key_values=self._past_key_values, use_cache=True)

        not_cached_q = outputs.logits

        self._prob_history = torch.cat([self._prob_history, not_cached_q], dim=1)

        last_q = not_cached_q[:, -1, :] #(batch_size, 1, vocab_size)
        self._past_key_values = outputs.past_key_values

        return last_q


    def _generate_with_kvcache(self,prefix:torch.Tensor,
                                    gamma:int,
                                    sample:Callable,
                                    **sampling_kwargs) -> torch.Tensor:
        """ forward the model gamma times

        Args:
            prefix (torch.Tensor): the prefix
            gamma (int): how many times approx guesses

        Returns:
            Torch.Tensor: prefix+generated tokens
        """
        x = prefix

        #prefill stage
        q = self.prefill(x)
        next_tok = sample(q, **sampling_kwargs)
        x = torch.cat((x, next_tok), dim=1)

        for _ in range(gamma):
            q = self.forward(x)
            next_tok = sample(q, **sampling_kwargs)
            x = torch.cat((x, next_tok), dim=1)

        self.clear()
        return x

    def decode_function(self, input_ids:torch.Tensor, length:int, return_text:bool=True) -> torch.Tensor:

        for step in range(length):
            with torch.no_grad():
                q = self.forward(input_ids)

            #print("prob history shape:", self._prob_history.shape)
            token_probabilities = torch.nn.functional.softmax(q, dim=-1)
            max_prob_index = torch.argmax(token_probabilities, dim=-1)
            next_token_id = max_prob_index

            # Append the sampled token to the input sequence
            input_ids = torch.cat([input_ids, next_token_id.unsqueeze(1)], dim=-1)

        all_probabilities = torch.nn.functional.softmax(self._prob_history[:, -length-1:-1, :], dim=-1).squeeze(0)

        if return_text:
            return input_ids[:, -length:], all_probabilities
        else:
            return all_probabilities

    @torch.no_grad()
    def generate(self, input:torch.Tensor, gamma:int, sample:Callable, **sampling_kwargs) -> torch.Tensor:
        output = self._generate_with_kvcache(input, gamma, sample, **sampling_kwargs)
        return output

    @torch.no_grad()
    def rollback_cache(self, end_pos : int):
        past_key_values_trimmed = []
        assert self._past_key_values
        for kv in self._past_key_values:
            k, v = kv
            # NOTE() the indexing is specific for bloom. This won't work for other models
            # For example llama k, v should be (batch, num_head, seq_len, hidden_dim)

            # Bloom is special one
            if isinstance(self._model, BloomForCausalLM):
                # k (batch * head, hidden_dim, seq); v (batch * head, seq, hidden_dim)
                k = k[:, :, :end_pos]
                v = v[:, :end_pos, :]
                kv_trimmed = (k, v)
                past_key_values_trimmed.append(kv_trimmed)
            else:
                # k, v (batch, head, seq, hidden_dim)
                k = k[:, :, :end_pos, :]
                v = v[:, :, :end_pos, :]
                kv_trimmed = (k, v)
                past_key_values_trimmed.append(kv_trimmed)

        self._past_key_values = past_key_values_trimmed
        self._prob_history = self._prob_history[:, :end_pos, :]

    def clear(self):
        self._past_key_values = None
        self._prob_history = None


from .load_data import LoadData
from .train_predacons import TrainPredacons
from .generate import Generate
from .data_preparation import DataPreparation
from .speculative_fast_generation import GPTFast
import torch
import pandas as pd

def rollout():
    print("Predacons rollout !!!")
    print("Predacons Version: v0.0.110")
    print("\nread_documents_from_directory -- Load data from directory")
    print("    directory -- Directory path")
    print("\nread_multiple_files -- Load data from multiple files")
    print("    file_paths -- list of File paths")
    print("\nclean_text -- Clean text")
    print("    text -- Text")
    print("\nread_csv -- Read csv file")
    print("    file_path -- File path")
    print("\ntrain_legacy -- Train Predacons")
    print("    train_file_path -- Train file path")
    print("    model_name -- Model name")
    print("    output_dir -- Output directory")
    print("    overwrite_output_dir -- Overwrite output directory")
    print("    per_device_train_batch_size -- Per device train batch size")
    print("    num_train_epochs -- Number of train epochs")
    print("    save_steps -- Save steps")
    print("    trust_remote_code -- Trust remote code")
    print("\ntrainer_legacy -- returns trainer")
    print("    train_file_path -- Train file path")
    print("    model_name -- Model name")
    print("    output_dir -- Output directory")
    print("    overwrite_output_dir -- Overwrite output directory")
    print("    per_device_train_batch_size -- Per device train batch size")
    print("    num_train_epochs -- Number of train epochs")
    print("    save_steps -- Save steps")
    print("    trust_remote_code -- Trust remote code")
    print("\ntrain -- Train Predacons")
    print("    train_file_path -- Train file path")
    print("    model_name -- Model name")
    print("    output_dir -- Output directory")
    print("    overwrite_output_dir -- Overwrite output directory")
    print("    per_device_train_batch_size -- Per device train batch size")
    print("    num_train_epochs -- Number of train epochs")
    print("    save_steps -- Save steps")
    print("    trust_remote_code -- Trust remote code")
    print("\ntrainer -- returns trainer")
    print("    train_file_path -- Train file path")
    print("    model_name -- Model name")
    print("    output_dir -- Output directory")
    print("    overwrite_output_dir -- Overwrite output directory")
    print("    per_device_train_batch_size -- Per device train batch size")
    print("    num_train_epochs -- Number of train epochs")
    print("    save_steps -- Save steps")
    print("    trust_remote_code -- Trust remote code")
    print("\ngenerate_text -- Generate text (Deprecating soon, use text_generate instead)")
    print("    model_path -- Model path")
    print("    sequence -- Sequence")
    print("    max_length -- Max length")
    print("    trust_remote_code -- Trust remote code (default False)")
    print("    use_fast_generation -- Use fast generation using speculative decoding (default False)")
    print("    draft_model_name -- Draft model name / path (default None)")
    print("\ngenerate_output -- returns output and tokenizer (Deprecating soon, use generate instead)")
    print("    model_path -- Model path")
    print("    sequence -- Sequence")
    print("    max_length -- Max length")
    print("    trust_remote_code -- Trust remote code (default False)")
    print("    use_fast_generation -- Use fast generation using speculative decoding (default False)")
    print("    draft_model_name -- Draft model name / path (default None)")
    print("\ngenerate -- Generate text or output")
    print("    model_path -- Model path")
    print("    sequence -- Sequence")
    print("    max_length -- Max length")
    print("    trust_remote_code -- Trust remote code (default False)")
    print("    use_fast_generation -- Use fast generation using speculative decoding (default False)")
    print("    draft_model_name -- Draft model name / path (default None)")
    print("    model -- give a preloaded Model (default None)")
    print("    tokenizer -- give a preloaded Tokenizer (default None)")
    print("\ntext_generate -- Generate text and print")
    print("    model_path -- Model path")
    print("    sequence -- Sequence")
    print("    max_length -- Max length")
    print("    trust_remote_code -- Trust remote code (default False)")
    print("    use_fast_generation -- Use fast generation using speculative decoding (default False)")
    print("    draft_model_name -- Draft model name / path (default None)")
    print("    model -- give a preloaded Model (default None)")
    print("    tokenizer -- give a preloaded Tokenizer (default None)")
    print("\nload_model -- Load model")
    print("    model_path -- Model path")
    print("    trust_remote_code -- Trust remote code (default False)")
    print("    use_fast_generation -- Use fast generation using speculative decoding (default False)")
    print("    draft_model_name -- Draft model name / path (default None)")
    print("\nload_tokenizer -- Load tokenizer")
    print("    tokenizer_path -- Tokenizer path")
    print("\ngenerate_text_data_source_openai -- Generate text data source using openai")
    print("    client -- openai client")
    print("    gpt_model -- GPT model used for generation")
    print("    prompt -- Prompt to generate data source")
    print("    number_of_examples -- Number of examples")
    print("    temperature -- Temperature (default 0.5)")
    print("\ngenerate_text_data_source_ll -- Generate text data source using local or hugging face llm")
    print("    model_path -- Model path or hugging face model name")
    print("    sequence -- Sequence (prompt)")
    print("    max_length -- Max length of the generated text")
    print("    number_of_examples -- Number of examples")
    print("    trust_remote_code -- Trust remote code")
    print("\nPredacons rollout !!!")

# Load data
def read_documents_from_directory(directory, encoding="utf-8"):
    """
    Read documents from a directory.

    Args:
        directory (str): The path to the directory containing the documents.
        encoding (str, optional): The encoding of the documents. Defaults to "utf-8".

    Returns:
        list: A list of documents read from the directory.
    """
    return LoadData.read_documents_from_directory(directory, encoding)

def read_multiple_files(file_paths):
    """
    Read and load data from multiple files.

    Args:
        file_paths (list): A list of file paths to read data from.

    Returns:
        object: The loaded data.

    """
    return LoadData.read_multiple_files(file_paths)

def clean_text(text):
    """
    Cleans the given text by removing any unwanted characters or formatting.

    Args:
        text (str): The text to be cleaned.

    Returns:
        str: The cleaned text.

    """
    return LoadData.clean_text(text)

def read_csv(file_path, encoding="utf-8"):
    """
    Read a CSV file and return the data as a pandas DataFrame.

    Parameters:
    - file_path (str): The path to the CSV file.
    - encoding (str, optional): The encoding of the CSV file. Default is 'utf-8'.

    Returns:
    - pandas.DataFrame: The data from the CSV file.

    Example:
    >>> data = read_csv('/path/to/file.csv')
    """
    return LoadData.read_csv(file_path, encoding)

# Train Predacons
def train_legacy(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,
          trust_remote_code = False,
          resume_from_checkpoint=True):
    """
    Trains the Predacons model using legacy training method.

    Args:
        train_file_path (str): The path to the training file.
        model_name (str): The name of the model.
        output_dir (str): The directory to save the trained model.
        overwrite_output_dir (bool): Whether to overwrite the output directory if it already exists.
        per_device_train_batch_size (int): The batch size for training.
        num_train_epochs (int): The number of training epochs.
        save_steps (int): The number of steps between saving checkpoints.
        trust_remote_code (bool, optional): Whether to trust remote code. Defaults to False.
        resume_from_checkpoint (bool, optional): Whether to resume training from a checkpoint. Defaults to True.
    """
    TrainPredacons.train(train_file_path = train_file_path,
        model_name = model_name,
        output_dir = output_dir,
        overwrite_output_dir = overwrite_output_dir,
        per_device_train_batch_size = per_device_train_batch_size,
        num_train_epochs = num_train_epochs,
        save_steps = save_steps,
        trust_remote_code = trust_remote_code,
        resume_from_checkpoint=resume_from_checkpoint)

# get trainer
def trainer_legacy(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,
          trust_remote_code = False):
    """
    Trains the Predacons model using legacy training method.

    Args:
        train_file_path (str): The path to the training file.
        model_name (str): The name of the model.
        output_dir (str): The directory to save the trained model.
        overwrite_output_dir (bool): Whether to overwrite the output directory if it already exists.
        per_device_train_batch_size (int): The batch size for training.
        num_train_epochs (int): The number of training epochs.
        save_steps (int): The number of steps between saving checkpoints.
        trust_remote_code (bool, optional): Whether to trust remote code. Defaults to False.
        resume_from_checkpoint (bool, optional): Whether to resume training from a checkpoint. Defaults to True.
    """
    return TrainPredacons.trainer(train_file_path=train_file_path,
        model_name = model_name,
        output_dir = output_dir,
        overwrite_output_dir = overwrite_output_dir,
        per_device_train_batch_size = per_device_train_batch_size,
        num_train_epochs = num_train_epochs,
        save_steps = save_steps,
        trust_remote_code = trust_remote_code)

def train(*args, **kwargs):
    return TrainPredacons.train(*args, **kwargs)

def trainer(*args, **kwargs):
    """
    Prepares a trainer instance for the Predacons model with the provided arguments and keyword arguments.
    
    This function serves as a wrapper that calls the `trainer` method of the `TrainPredacons` class, forwarding all
    received arguments and keyword arguments. It is designed to configure and return a trainer instance without
    immediately starting the training process, allowing for further customization or inspection of the trainer
    configuration before training.
    
    ## Parameters

    - `*args`: Arbitrary positional arguments. Currently, this method does not utilize positional arguments but is designed to be flexible for future extensions.

    - `**kwargs`: Arbitrary keyword arguments used for configuring the training process. The supported keywords include:

    - `use_legacy_trainer` (bool): If `True`, uses a legacy training approach. Default is `False`.
    - `model_name` (str): The name or path of the pre-trained model to be used.
    - `train_file_path` (str): Path to the training dataset file.
    - `tokenizer` (Tokenizer): An instance of a tokenizer.
    - `output_dir` (str): Directory where the model and tokenizer will be saved after training.
    - `overwrite_output_dir` (bool): If `True`, overwrite the output directory.
    - `per_device_train_batch_size` (int): Batch size per device during training.
    - `num_train_epochs` (int): Total number of training epochs.
    - `quantization_config` (dict): Configuration for model quantization.
    - `auto_quantize` (str): Automatically apply quantization. Accepts "4bit"/"high" for high compression or "8bit"/"low" for lower compression.
    - `trust_remote_code` (bool): If `True`, allows the execution of remote code during model loading.
    - `peft_config` (dict): Configuration for Parameter Efficient Fine-Tuning (PEFT) techniques like LoRA.
    - `auto_lora_config` (bool): If `True`, automatically configures LoRA for the model.
    - `training_args` (TrainingArguments): Configuration arguments for the Hugging Face `Trainer`.
    - `train_file_type` (str): Type of the training file. Supported types are "text", "csv", "json".
    - `train_dataset` (Dataset): A pre-loaded dataset. If provided, `train_file_path` is ignored.
    - `preprcess_function` (callable): A function to preprocess the dataset.
    - `resume_from_checkpoint` (str): Path to a directory containing a checkpoint from which training is to resume.
    - `save_steps` (int): Number of steps after which the model is saved.

    ## Returns

    - Returns an instance of `Trainer` or `SFTTrainer`, configured according to the provided arguments. This object is ready to be used for training the model.

    ## Example Usage

    ```python
    trainer = TrainPredacons.trainer(
        model_name='bert-base-uncased',
        train_file_path='./data/train.txt',
        tokenizer=my_tokenizer,
        output_dir='./model_output',
        num_train_epochs=3,
        per_device_train_batch_size=8,
        auto_quantize='4bit'
    )
    """
    return TrainPredacons.trainer(*args, **kwargs)

# Generate text
def generate_text(model_path, sequence, max_length,trust_remote_code = False,use_fast_generation=False, draft_model_name=None):
    """
    Generate text using the specified model.

    Args:
        model_path (str): The path to the model.
        sequence (str): The input sequence to generate text from.
        max_length (int): The maximum length of the generated text.
        trust_remote_code (bool, optional): Whether to trust remote code. Defaults to False.
        use_fast_generation (bool, optional): Whether to use fast generation. Defaults to False.
        draft_model_name (str, optional): The name of the draft model. Defaults to None.

    Returns:
        str: The generated text.
    """
    print("For repetitive generation first load model and then use text_generate. It will be faster.")
    print("will be deprecated soon, use text_generate")
    if use_fast_generation:
        print("generate_text using fast generation")
        if draft_model_name == None:
            print("Draft model is required for fast generation. Using base model as draft model, but it may increase memory utilization. try to use draft model name for better performance.")
            draft_model_name = model_path
        return GPTFast.generate_text_fast(model_path, draft_model_name, sequence, max_length,trust_remote_code = trust_remote_code)
    else:
        print("generate_text using default generation")
        return Generate.generate_text(model_path, sequence, max_length,trust_remote_code = trust_remote_code) 

# Generate output
def generate_output(model_path, sequence, max_length,trust_remote_code = False,use_fast_generation=False, draft_model_name=None):
    """
    Generates output using the specified model.

    Args:
        model_path (str): The path to the model.
        sequence (str): The input sequence for generating the output.
        max_length (int): The maximum length of the generated output.
        trust_remote_code (bool, optional): Whether to trust remote code. Defaults to False.
        use_fast_generation (bool, optional): Whether to use fast generation. Defaults to False.
        draft_model_name (str, optional): The name of the draft model. Defaults to None.

    Returns:
        str: The generated output.
    """
    print("For repetitive generation first load model and then use generate. It will be faster.")
    print("will be deprecated soon, use generate instead")
    if use_fast_generation:
        print("generate_output using fast generation")
        if draft_model_name == None:
            print("Draft model is required for fast generation. Using base model as draft model, but it may increase memory utilization. try to use draft model name for better performance.")
            draft_model_name = model_path
        return GPTFast.generate_output_fast(model_path, draft_model_name, sequence, max_length,trust_remote_code = trust_remote_code)
    else:
        print("generate_output using default generation")
        return Generate.generate_output(model_path, sequence, max_length,trust_remote_code = trust_remote_code) 

# generate new

def generate(*args, **kwargs):
    """
    Generates output based on the provided arguments.

    Args:
        *args: Variable length arguments.
        **kwargs: Keyword arguments.

    Keyword Args:
        model_path (str): The path to the model file.
        sequence (str): The input sequence to generate output from.
        max_length (int, optional): The maximum length of the generated output. Defaults to 50.
        trust_remote_code (bool, optional): Whether to trust remote code. Defaults to False.
        use_fast_generation (bool, optional): Whether to use fast generation. Defaults to False.
        draft_model_name (str, optional): The name of the draft model. Defaults to None.
        model (object): The model object.
        tokenizer (object): The tokenizer object.

    Returns:
        str: The generated output.

    Raises:
        ValueError: If the arguments are invalid.
    """
    if 'model_path' in kwargs and 'sequence' in kwargs:
        model_path = kwargs['model_path']
        sequence = kwargs['sequence']
        max_length = kwargs.get('max_length', 50)
        trust_remote_code = kwargs.get('trust_remote_code', False)
        use_fast_generation = kwargs.get('use_fast_generation', False)
        draft_model_name = kwargs.get('draft_model_name', None)
        if use_fast_generation:
            print("generate_output using fast generation")
            if draft_model_name == None:
                print("Draft model is required for fast generation. Using base model as draft model, but it may increase memory utilization. try to use draft model name for better performance.")
                draft_model_name = model_path
            return GPTFast.generate_output_fast(model_path, draft_model_name, sequence, max_length,trust_remote_code = trust_remote_code)
        else:
            print("generate_output using default generation")
            return Generate.generate_output(model_path, sequence, max_length,trust_remote_code = trust_remote_code) 
    
    elif 'model' in kwargs and 'tokenizer' in kwargs and 'sequence' in kwargs:
        model = kwargs['model']
        tokenizer = kwargs['tokenizer']
        sequence = kwargs['sequence']
        max_length = kwargs.get('max_length', 50)
        trust_remote_code = kwargs.get('trust_remote_code', False)
        try:
            if type(model) == torch._dynamo.eval_frame.OptimizedModule:
                print("generate_output using fast generation")
                return GPTFast.generate_output_from_model(model, tokenizer, sequence, max_length)
            else:
                return Generate.generate_output_from_model(model, tokenizer, sequence, max_length,trust_remote_code=trust_remote_code)
        except Exception as e:
            print("Exception occurred while loading torch._dynamo.eval_frame.OptimizedModule")
            print("generate_output using default generation")
            return Generate.generate_output_from_model(model, tokenizer, sequence, max_length,trust_remote_code=trust_remote_code)
    else:
        raise ValueError("Invalid arguments")
    
def text_generate(*args, **kwargs):
    """
    Generate text using the specified arguments.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        str: The generated text.

    """
    output, tokenizer = generate(*args, **kwargs)
    print(tokenizer.decode(output[0], skip_special_tokens=True))
    return tokenizer.decode(output[0], skip_special_tokens=True)

# Data preparation
def generate_text_data_source_openai(client,gpt_model,prompt,number_of_examples,temperature =0.5):
    """
    Generates a text data source using OpenAI's GPT model.

    Parameters:
    - client: The OpenAI client object.
    - gpt_model: The GPT model to use for generating text.
    - prompt: The prompt to start the text generation.
    - number_of_examples: The number of text examples to generate.
    - temperature: The temperature parameter for controlling the randomness of the generated text. Default is 0.5.

    Returns:
    - The generated text data source.
    """
    return DataPreparation.generate_text_data_source_openai(client,gpt_model,prompt,number_of_examples,temperature)

def generate_text_data_source_llm(model_path, sequence, max_length,number_of_examples,trust_remote_code=False):
    """
    Generate a text data source for language model training.

    Args:
        model_path (str): The path to the language model.
        sequence (str): The input sequence to generate data from.
        max_length (int): The maximum length of the generated data.
        number_of_examples (int): The number of examples to generate.
        trust_remote_code (bool, optional): Whether to trust remote code. Defaults to False.

    Returns:
        str: The generated text data source.
    """
    return DataPreparation.generate_text_data_source_llm(model_path, sequence, max_length,number_of_examples,trust_remote_code=trust_remote_code)

# Get model and tokenizer
def load_model(model_path,trust_remote_code=False,use_fast_generation=False, draft_model_name=None):
    """
    Load a model from the specified model_path.

    Args:
        model_path (str): The path to the model.
        trust_remote_code (bool, optional): Whether to trust remote code. Defaults to False.
        use_fast_generation (bool, optional): Whether to use fast generation. Defaults to False.
        draft_model_name (str, optional): The name of the draft model. Defaults to None.

    Returns:
        Model: The loaded model.

    Raises:
        FileNotFoundError: If the model_path does not exist.
    """
    if use_fast_generation:
        print("load_model using fast generation")
        if draft_model_name == None:
            print("Draft model is required for fast generation. Using base model as draft model, but it may increase memory utilization. try to use draft model name for better performance.")
            draft_model_name = model_path
        return GPTFast.load_model(model_path, draft_model_name,trust_remote_code=trust_remote_code)
    else:
        print("load_model using default generation")
        return Generate.load_model(model_path,trust_remote_code=trust_remote_code)

def load_tokenizer(tokenizer_path):
    """
    Loads a tokenizer from the specified path.

    Args:
        tokenizer_path (str): The path to the tokenizer file.

    Returns:
        Tokenizer: The loaded tokenizer object.
    """
    return Generate.load_tokenizer(tokenizer_path)
 




from transformers import TextDataset, DataCollatorForLanguageModeling,BitsAndBytesConfig
from transformers import AutoTokenizer,AutoModelForPreTraining,AutoModelForCausalLM
from transformers import Trainer, TrainingArguments
from peft import LoraConfig
from datasets import load_dataset
import torch
from trl import SFTTrainer

class TrainPredacons:
    def get_feature_list(train_dataset):
        feature_list  = []
        for feature in  list(train_dataset['train'].features):
            if feature not in ['input_ids', 'attention_mask', 'Unnamed: 0']:
                    feature_list.append(feature)
        print(feature_list)
        return feature_list
    
    def create_formatting_func(train_dataset):
        feature_list  = TrainPredacons.get_feature_list(train_dataset)
        def auto_formatting_func(example):
            func_str =''
            for feature in feature_list:
                feature_str = f"{feature}: {example[feature][0]}\n"
                func_str = func_str + feature_str
            text = f"{func_str}"
            return [text]

        return auto_formatting_func

    def formatting_func(example):
        text = f"{example['text'][0]}"
        return [text]

    def __load_dataset(file_path, tokenizer, block_size = 128):
        dataset = TextDataset(
            tokenizer = tokenizer,
            file_path = file_path,
            block_size = block_size,
        )
        return dataset
    def lld(file_path, tokenizer, block_size = 128):
        dataset = TextDataset(
            tokenizer = tokenizer,
            file_path = file_path,
            block_size = block_size,
        )
        return dataset
    def __load_data_collator(tokenizer, mlm = False):
        data_collator = DataCollatorForLanguageModeling(
            tokenizer=tokenizer,
            mlm=mlm,
        )
        return data_collator


    def __trainer(*args,**kwargs):

        # use_legacy_trainer
        print("Initiating predacons training...")
        use_legacy_trainer = kwargs.get('use_legacy_trainer', False)
        if(use_legacy_trainer):
            print("Using lagacy trainer will be removed in further release.")
            tokenizer = AutoTokenizer.from_pretrained(kwargs.get('model_name', None))
            train_dataset = TrainPredacons.__load_dataset(kwargs.get('train_file_path', None), kwargs.get('tokenizer', None))
            data_collator = TrainPredacons.__load_data_collator(kwargs.get('tokenizer', None))

            tokenizer.save_pretrained(kwargs.get('output_dir', None))
            model = None
            try:
                model = AutoModelForPreTraining.from_pretrained(kwargs.get('model_name', None),trust_remote_code=kwargs.get('trust_remote_code', None))
            except:
                model = AutoModelForCausalLM.from_pretrained(kwargs.get('model_name', None),trust_remote_code=kwargs.get('trust_remote_code', None))

            model.save_pretrained(kwargs.get('output_dir', None))

            training_args = TrainingArguments(
                    output_dir=kwargs.get('output_dir', None),
                    overwrite_output_dir=kwargs.get('overwrite_output_dir', None),
                    per_device_train_batch_size=kwargs.get('per_device_train_batch_size', None),
                    num_train_epochs=kwargs.get('num_train_epochs', None),
                )

            trainer = Trainer(
                    model=model,
                    args=training_args,
                    data_collator=data_collator,
                    train_dataset=train_dataset,
            )
            return trainer

        # load model and tokenizer
        print("Loading model and tokenizer...")
        model_name = ""
        model = None
        tokenizer = None
        if 'model_name' in kwargs:
            model_name = kwargs['model_name']
            quantization_config = None
            if 'quantization_config' in kwargs:
                quantization_config = kwargs['quantization_config']
                print("Using user provided quantization_config...")
            auto_quantize = kwargs.get('auto_quantize', "")

            if auto_quantize.lower() == "4bit" or auto_quantize.lower() == "high":
                print("Using 4bit quantization...")
                quantization_config = BitsAndBytesConfig(
                    load_in_4bit=True,
                    bnb_4bit_quant_type="nf4",
                    bnb_4bit_compute_dtype=torch.bfloat16
                )
            elif  auto_quantize.lower() == "8bit" or auto_quantize.lower() == "low":
                print("Using 8bit quantization...")
                quantization_config = BitsAndBytesConfig(
                    load_in_8bit =True,
                    bnb_4bit_compute_dtype=torch.bfloat16
                )
            else:
                print("Error: Either give 4bit/high or 8bit/low as the value of auto_quantize. moving ahead without any quantization" )
            tokenizer = AutoTokenizer.from_pretrained(model_name)
            trust_remote_code = kwargs.get('trust_remote_code', False)

            model = AutoModelForCausalLM.from_pretrained(model_name,quantization_config = quantization_config,trust_remote_code =trust_remote_code)
            print(f"Model ready: {model_name} m")

        elif ('model' in kwargs and 'tokenizer' in kwargs):
            print("Using provided model and tokenizer...")
            try:
                model = kwargs['model']
                tokenizer = kwargs['tokenizer']
            except Exception as e:
                raise ValueError(f"error while loading model and tokenizer {e}")
        else:
            raise Exception("provide model name or model and tokenizer")

        # get output dir
        output_dir =""
        if 'output_dir' in kwargs:
            output_dir =  kwargs['output_dir']
            tokenizer.save_pretrained(output_dir)
            model.save_pretrained(output_dir)
        else:
            print("Warning: Output directory not provided saving everything to root dir, better stop execution and provide value in output_dir... " )

        # load peft_config
        peft_config = None
        auto_lora_config = kwargs.get('auto_lora_config',False)
        if auto_lora_config == True:
            peft_config = LoraConfig(
                r = 8,
                target_modules = ["q_proj", "o_proj", "k_proj", "v_proj",
                                "gate_proj", "up_proj", "down_proj"],
                task_type = "CAUSAL_LM",
            )
        else:
            peft_config = kwargs.get('peft_config',None)
        print(f"peft_config: {peft_config}")

        # load training_args
        print("Loading training_args...")
        training_args = kwargs.get('training_args',None)
        if training_args == None:
            per_device_train_batch_size = kwargs.get('per_device_train_batch_size',None)
            num_train_epochs =  kwargs.get('num_train_epochs',None)
            training_args = TrainingArguments(
                    output_dir = output_dir,
                    overwrite_output_dir = kwargs.get('overwrite_output_dir',None),
                    per_device_train_batch_size = kwargs.get('per_device_train_batch_size',1),
                    num_train_epochs=num_train_epochs,
                    fp16=True if auto_lora_config == True else True ,
                    optim="paged_adamw_8bit" if auto_lora_config == True else "paged_adamw_8bit",
                    save_steps = kwargs.get('save_steps',500)
                )

        # dataset configuration
        print(f"Creating data set...")
        train_file_type = kwargs.get('train_file_type',None)
        train_file_path = kwargs.get('train_file_path',None)
        train_dataset = kwargs.get('train_dataset',None)
        preprcess_function = kwargs.get('preprcess_function',None)
        if train_file_path != None and train_dataset == None:
            if train_file_type not in ("text","csv","json"):
                print("Error: File type not supported use text,csv or json trying to process as text")
                train_file_type = "text"
            train_dataset = load_dataset(train_file_type, data_files={"train": train_file_path})
            if train_file_type == "text":
                train_dataset = train_dataset.map(lambda samples: tokenizer(samples["text"]), batched=True)
            else:
                if preprcess_function == None:
                    print("No preprcess_function provided using default preprcess_function")
                    first_feature = TrainPredacons.get_feature_list(train_dataset)[0]
                    print(f"Using first_feature: {first_feature}, pass preprcess_function to use a custom preprocess_function")
                    train_dataset = train_dataset.map(lambda samples: tokenizer(samples[first_feature]), batched=True)
                else:
                    train_dataset = train_dataset.map(preprcess_function, batched=True)
        else:
            print("Please provide train_file_path to load dataset other sources will be added in next release")
        print(f"train_dataset: {train_dataset}")
        try:
            print("Using custom formatting function...")
            formatting_func = TrainPredacons.create_formatting_func(train_dataset)
        except:
            print("Using default formatting function...")
            formatting_func = TrainPredacons.formatting_func
        trainer = SFTTrainer(
            model=model,
            train_dataset=train_dataset["train"],
            args=training_args,
            peft_config=peft_config,
            formatting_func = formatting_func

        )
        return trainer

    def __train(*args,**kwargs):

        trainer = TrainPredacons.__trainer(*args,**kwargs)
        try:
            checkpoint_folder_exists = any(folder.startswith('checkpoint') for folder in os.listdir(kwargs.get('output_dir', "")))
            trainer.train(resume_from_checkpoint = (checkpoint_folder_exists and kwargs.get('resume_from_checkpoint', "")))
        except:
            print('Error: Failed to resume from checkpoint. training from scratch.')
            trainer.train()
        trainer.save_model()

    def trainer(*args,**kwargs):
        return TrainPredacons.__trainer(*args,**kwargs)

    def train(*args,**kwargs):
        return TrainPredacons.__train(*args,**kwargs)